﻿using System;
using System.Threading;

using wclCommon;
using wclBluetooth;
using System.Collections.Generic;

namespace GattClient
{
    internal class WorkingThread
    {
        // Termination event.
        private ManualResetEvent FEvent;
        // Connected devices list.
        private List<Int64> FDevices;
        private wclBluetoothManager FManager;
        private wclBluetoothLeBeaconWatcher FWatcher;
        private wclBluetoothRadio FRadio;
        private Thread FThread;

        // The event fires when Bluetooth Manager opened.
        private void FManager_AfterOpen(object sender, EventArgs e)
        {
            // Try to find working radio.
            for (Int32 i = 0; i < FManager.Count; i++)
            {
                if (FManager[i].Available && FManager[i].Api == wclBluetoothApi.baMicrosoft)
                {
                    FRadio = FManager[i];
                    break;
                }
            }

            if (FRadio == null)
                Trace("No Bluetooth radio found.");
            else
            {
                // Try to start watcher.
                Int32 Res = FWatcher.Start(FRadio);
                if (Res != wclErrors.WCL_E_SUCCESS)
                    Trace("Start beacon watcher failed.", Res);
                else
                    Trace("Devices monitoring started.");
            }
        }

        private void FWatcher_OnAdvertisementFrameInformation(Object Sender, Int64 Address, Int64 Timestamp, SByte Rssi,
            String Name, wclBluetoothLeAdvertisementType PacketType, wclBluetoothLeAdvertisementFlag Flags)
        {
            // Advertisement frame received. Check device's name.
            if (Name != "" && Name.Contains("iBesy-button"))
            {
                // Make sure device is not connected yet.
                if (!FDevices.Contains(Address))
                {
                    Trace("Device found. Name: " + Name + ". Address: " + Address.ToString("X12"));
                    // Start connection to device.
                    FDevices.Add(Address);
                    Client client = new Client(this, FRadio, Address);
                }
            }
        }

        // The thread procedure.
        private void Proc()
        {
            FDevices = new List<Int64>();

            // Create BluetoothManager and assign event handlers.
            FManager = new wclBluetoothManager();
            // We need only this event.
            FManager.AfterOpen += FManager_AfterOpen;
            // Create BeaconWatcher and Assign event handlers.
            FWatcher = new wclBluetoothLeBeaconWatcher();
            // We need only this event.
            FWatcher.OnAdvertisementFrameInformation += FWatcher_OnAdvertisementFrameInformation;

            // Try to open Bluetooth Manager.
            Int32 Res = FManager.Open();
            if (Res != wclErrors.WCL_E_SUCCESS)
                Trace("Failed to open Bluetooth Manager.", Res);

            // Wait for termination.
            IntPtr[] Handlers = { FEvent.SafeWaitHandle.DangerousGetHandle() };
            // Process APC until thread terminated.
            while (wclHelpers.WaitEx(Handlers, wclWinAPI.INFINITE) != wclWinAPI.WAIT_OBJECT_0)
                ;

            // Close Bluetooth Manager. It forces to close all connected clients and stop beacons watcher.
            FEvent.Close();
            FEvent = null;
            FManager.Close();
            FManager = null;
            FWatcher = null;
            FRadio = null;
            FDevices = null;
        }

        public WorkingThread()
        {
            FManager = null;
            FWatcher = null;
            FRadio = null;
            FThread = null;
            FEvent = new ManualResetEvent(false);

            // Create and start working thread.
            FThread = new Thread(Proc);
            FThread.Start();
        }

        // Terminates the thread.
        public void Terminate()
        {
            FEvent.Set();
            FThread.Join();
            FThread = null;
        }

        // Called by the client class to remove the device from the connected devices list.
        public void RemoveDevice(Int64 Address)
        {
            if (FDevices != null)
                FDevices.Remove(Address);
        }

        // Writes message to the console.
        public void Trace(String Msg)
        {
            Console.WriteLine(Msg);
        }

        // Write error message to the console.
        public void Trace(String Msg, Int32 Error)
        {
            Trace(Msg + " Error: 0x" + Error.ToString("X8"));
        }
    }
}
